#include <NTL/config.h>
#include <NTL/GF2X.h>

#ifndef MUL_KARA_THRESHOLD
#define MUL_KARA_THRESHOLD 2
#endif

#ifndef MUL_TOOM_THRESHOLD
#define MUL_TOOM_THRESHOLD 3
#endif

#undef NTL_GF2X_ALTCODE1
#undef NTL_GF2X_ALTCODE
#define NTL_GF2X_NOINLINE

/* taken from GF2X.c */
#if (defined(NTL_GF2X_ALTCODE1))

#define NTL_EFF_BB_MUL_CODE0 NTL_ALT1_BB_MUL_CODE0
#define NTL_EFF_BB_MUL_CODE1 NTL_ALT1_BB_MUL_CODE1
#define NTL_EFF_BB_MUL_CODE2 NTL_ALT1_BB_MUL_CODE2
#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_ALT1_SHORT_BB_MUL_CODE1
#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_ALT1_HALF_BB_MUL_CODE0

#elif (defined(NTL_GF2X_ALTCODE))

#define NTL_EFF_BB_MUL_CODE0 NTL_ALT_BB_MUL_CODE0
#define NTL_EFF_BB_MUL_CODE1 NTL_ALT_BB_MUL_CODE1
#define NTL_EFF_BB_MUL_CODE2 NTL_ALT_BB_MUL_CODE2
#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_ALT_SHORT_BB_MUL_CODE1
#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_ALT_HALF_BB_MUL_CODE0

#else

#define NTL_EFF_BB_MUL_CODE0 NTL_BB_MUL_CODE0
#define NTL_EFF_BB_MUL_CODE1 NTL_BB_MUL_CODE1
#define NTL_EFF_BB_MUL_CODE2 NTL_BB_MUL_CODE2
#define NTL_EFF_SHORT_BB_MUL_CODE1 NTL_SHORT_BB_MUL_CODE1
#define NTL_EFF_HALF_BB_MUL_CODE0 NTL_HALF_BB_MUL_CODE0

#endif

static 
void mul1(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
{
NTL_EFF_BB_MUL_CODE0
}

#ifdef NTL_GF2X_NOINLINE
  #define mul1_IL mul1
#else

static inline
void mul1_inline(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
{


NTL_EFF_BB_MUL_CODE0


}

#define mul1_IL mul1_inline 
#endif

static 
void Mul1(_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)
{
NTL_EFF_BB_MUL_CODE1
}

static 
void AddMul1(_ntl_ulong *cp, const _ntl_ulong* bp, long sb, _ntl_ulong a)
{
NTL_EFF_BB_MUL_CODE2
}

static
void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
{
   _ntl_ulong hs0, hs1;
   _ntl_ulong hl2[2];

   hs0 = a[0] ^ a[1];
   hs1 = b[0] ^ b[1];

   mul1_IL(c, a[0], b[0]);
   mul1_IL(c+2, a[1], b[1]);
   mul1_IL(hl2, hs0, hs1);


   hl2[0] = hl2[0] ^ c[0] ^ c[2];
   hl2[1] = hl2[1] ^ c[1] ^ c[3];

   c[1] ^= hl2[0];
   c[2] ^= hl2[1];
}

/* an = bn = n > 0 */
static inline void
mul_basecase_n (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                long n)
{
  long i;

  switch (n)
    {
    case 1:
      mul1 (c, a[0], b[0]);
      return;
    case 2:
      mul2 (c, a, b); // faster than default
      return;
    default:
      Mul1 (c, a, n, b[0]);
      for (i = 1; i < n; i++)
        AddMul1 (c + i, a, n, b[i]);
      return;
    }
}

static
void KarMul(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, 
            long len, _ntl_ulong *stk)
{
   if (len < MUL_KARA_THRESHOLD)
     {
       mul_basecase_n (c, a, b, len);
       return;
     }

   long ll, lh, i, ll2, lh2;
   const _ntl_ulong *a0, *a1, *b0, *b1;
   _ntl_ulong *a01, *b01, *h;

   lh = len >> 1;
   ll = (len+1) >> 1;

   ll2 = ll << 1;
   lh2 = lh << 1;

   a01 = stk;  stk += ll+1;
   b01 = stk;  stk += ll+1;
   h = stk; stk += ll2+1;

   a0 = a;
   a1 = a+ll;
   b0 = b;
   b1 = b+ll;

   KarMul(c, a0, b0, ll, stk);
   KarMul(c+ll2, a1, b1, lh, stk);

   for (i = 0; i < lh; i++) {
      a01[i] = a[i] ^ a[i+ll];
      b01[i] = b[i] ^ b[i+ll];
   }

   if (lh < ll) {
      a01[lh] = a[lh];
      b01[lh] = b[lh];
   }

   KarMul(h, a01, b01, ll, stk);

   for (i = 0; i < ll2; i++)
      h[i] ^= c[i];

   for (i = 0; i < lh2; i++)
      h[i] ^= c[i+ll2];

   for (i = 0; i < ll2; i++)
      c[i+ll] ^= h[i];
}

/* c <- a + b */
static
void Add (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, long n)
{
  long i;
  for (i = 0; i < n; i++)
    c[i] = a[i] ^ b[i];
}

/* c <- c + a + b */
static
void Add3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, long n)
{
  long i;
  for (i = 0; i < n; i++)
    c[i] ^= a[i] ^ b[i];
}

/* c <- a + x * b, return carry out */
static
_ntl_ulong AddLsh1 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 1) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 1);
      c[i] = t;
    }
  return cy;
}

/* c <- a + x^4 * b, return carry out */
static
_ntl_ulong AddLsh4 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 4) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 4);
      c[i] = t;
    }
  return cy;
}

/* c <- x * c, return carry out */
static
_ntl_ulong Lsh1 (_ntl_ulong *c, long n, _ntl_ulong cy)
{
  _ntl_ulong t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = (c[i] << 1) | cy;
      cy = c[i] >> (NTL_BITS_PER_LONG - 1);
      c[i] = t;
    }
  return cy;
}

/* c <- c/x, return carry */
static
_ntl_ulong Rsh1 (_ntl_ulong *c, long n)
{
  _ntl_ulong cy = 0, t;
  long i;
  for (i = n - 1; i >= 0; i--)
    {
      t = (c[i] >> 1) | (cy << (NTL_BITS_PER_LONG - 1));
      cy = c[i] & 1UL;
      c[i] = t;
    }
#ifdef DEBUG
  if (cy) abort();
#endif
  return cy;
}

/* c <- a + cy, return carry out (0 for n > 0, cy for n=0) */
static
_ntl_ulong Add1 (_ntl_ulong *c, const _ntl_ulong *a, long n, _ntl_ulong cy)
{
  if (n)
    {
      long i;
      c[0] = a[0] ^ cy;
      for (i = 1; i < n; i++)
        c[i] = a[i];
      return 0;
    }
  else
    return cy;
}

/* let c = q*(1+x^2) + X^n*r with X = x^NTL_BITS_PER_LONG and deg(r) < 2
   then c <- q, returns r.
   (Algorithm from Michel Quercia.)
*/
static _ntl_ulong DivOnePlusX2 (_ntl_ulong *c, long n)
{
  _ntl_ulong t = 0;
  long i;

  for (i = 0; i < n; i++) {
    t ^= c[i];
    t ^= t << 2;
    t ^= t << 4;
    t ^= t << 8;
    t ^= t << 16;
#if (NTL_BITS_PER_LONG == 64)
    t ^= t << 32;
#elif (NTL_BITS_PER_LONG != 32)
#error "NTL_BITS_PER_LONG should be 32 or 64"
#endif
    c[i] = t;
    t >>= (NTL_BITS_PER_LONG - 2);
  }
  return t;
}

/*
  c := proc(t) c0+t*c1+t^2*c2+t^3*c3+t^4*c4 end:
  solve ({c(1) = ab1, c(x) = abx, expand(x^4*c(1/x)) = abix}, 
          {c1, c2, c3});
  After ab1 <- ab1 - c0 - c4,
        abx <- (abx - c0 - c4*x^4)/x
        abix <- (abix - c0*x^4 - c4)/x
  we have:
     B := c1*x^2 + c2*x + c3:
     C := c1 + c2 + c3:
     D := c1 + c2*x + c3*x^2:
  solve({C=c1+c2+c3, D=c1+c2*x+c3*x^2, B=c1*x^2+c2*x+c3}, {c1, c2, c3});
  map(proc(x) lhs(x)=evala(Factor(numer(rhs(x)) mod 2))/
       evala(Factor(denom(rhs(x)) mod 2)) end, %);
  t2 := (B + D) / (1+x)^2:          # c1+c3
  c2 := C + t2                      # c2
  c1 := (x*c2 + B + t2)/(1+x)^2:    # c1
  c3 = c1 + t2                      # c3
 */
static
void ToomMul (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
              long n, _ntl_ulong *stk)
{
  while (n && a[n-1] == 0 && b[n-1] == 0)
    {
      c[2*n-1] = 0;
      c[2*n-2] = 0;
      n --;
    }

  if (n < MUL_TOOM_THRESHOLD)
    return KarMul (c, a, b, n, stk);

  long k = (n + 2) / 3; /* ceil(n/3) */
  long r = n - 2 * k;
  _ntl_ulong cy;
  _ntl_ulong *A = c;
  _ntl_ulong *B = stk + 2 * k + 1;
  _ntl_ulong *C = c + 2 * k;
  _ntl_ulong *D = stk;
  _ntl_ulong *E = c + 4 * k;

  cy = AddLsh1 (c, a + k, a + 2 * k, r);   /* a1 + x a2 */
  cy = Add1 (c + r, a + k + r, k - r, cy);
  c[k] = (cy << 1) ^ AddLsh1 (c, a, c, k); /* a0 + x a1 + x^2 a2 */
  cy = AddLsh1 (c + k + 1, b + k, b + 2 * k, r);
  cy = Add1 (c + k + r + 1, b + k + r, k - r, cy);
  c[2 * k + 1] = (cy << 1) ^ AddLsh1 (c + k + 1, b, c + k + 1, k);
  ToomMul (D, c, c + k + 1, k + 1, stk + 2 * k + 2); /* D : 2*k+1 */

  c[k] = AddLsh1 (c, a + k, a, k);   /* a1 + x a0 */
  cy = AddLsh1 (c, a + 2 * k, c, r); /* x^2 a0 + x a1 + a2 */
  c[k] = (c[k] << 1) ^ Lsh1 (c + r, k - r, cy);
  c[2 * k + 1] = AddLsh1 (c + k + 1, b + k, b, k); /* b1 + x b0 */
  cy = AddLsh1 (c + k + 1, b + 2 * k, c + k + 1, r);
  c[2 * k + 1] = (c[2 * k + 1] << 1) ^ Lsh1 (c + k + r + 1, k - r, cy);
  ToomMul (B, c, c + k + 1, k + 1, stk + 4 * k + 3); /* B : 2*k+1 */

  Add (c, a, a + k, k);
  Add (c, c, a + 2 * k, r);
  Add (stk + 4*k+2, b, b + k, k);
  Add (stk + 4*k+2, stk + 4*k+2, b + 2 * k, r);
  ToomMul (C, c, stk + 4*k+2, k, stk + 5*k+2); /* C : 2*k */

  ToomMul (A, a, b, k, stk+4*k+2); /* A : 2*k */

  ToomMul (E, a + 2 * k, b + 2 * k, r, stk+4*k+2); /* E : 2*r */

  /* C <- C - A - E  = c1 + c2 + c3 */
  Add3 (C, A, E, 2 * r);
  Add (C + 2 * r, C + 2 * r, A + 2 * r, 2 * (k - r));

  /* B <- (B - A*x^4 - E)/x */
  Add (B, B, E, 2 * r);
  B[2*k] ^= AddLsh4 (B, B, A, 2 * k);
  Rsh1 (B, 2*k+1); /* (a0*b1+a1*b0)*x^2+(a0*b2+a1*b1+a2*b0)*x+a1*b2+a2*b1 */

  /* D <- (D - A - E*x^4)/x */
  Add (D, D, A, 2 * k);
  D[2*r] ^= AddLsh4 (D, D, E, 2 * r);
  Rsh1 (D, 2*k+1); /* (a1*b2+a2*b1)*x^2+(a0*b2+a1*b1+a2*b0)*x+a0*b1+a1*b0 */

  /* D <- (B + D)/(1+x)^2 */
  Add (D, D, B, 2*k+1); /* (a1*b2+a2*b1+a0*b1+a1*b0)*(x^2+1) */
  DivOnePlusX2 (D, 2*k+1);
  //  if (D[2*k]) abort();

  /* C <- C + t2 */
  Add (C, C, D, 2*k);

  /* B <- (B + t2 + x*c2) / (1+x)^2 */
  Add (B, B, D, 2*k);
  B[2*k] ^= AddLsh1 (B, B, C, 2*k);
  DivOnePlusX2 (B, 2*k+1);
  //  if (B[2*k]) abort();

  Add3 (c + 3 * k, B, D, 2 * k);
  Add (c + k, c + k, B, 2 * k);
}

