#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/resource.h>

#include "mul.c"

#define N 2000000

int
cputime ()
{
  struct rusage rus;

  getrusage (0, &rus);
  return rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
}

unsigned long KarMulMem (unsigned long n)
{
  unsigned long hn, sp;

  sp = 0;
  while (n > 1)
    {
      hn = (n + 1) >> 1;
      sp += (hn << 2) + 3;
      n = hn;
    }
  return sp;
}

int
main (int argc, char *argv[])
{
  unsigned long n;
  unsigned long i, *c, *a, *b, *stk;
  int st;
  int nokara = 0;

  if (strcmp (argv[1], "-nokara") == 0)
    {
      nokara = 1;
      argc --;
      argv ++;
    }

  n = atoi (argv[1]);

  a = (unsigned long*) malloc (n * sizeof(unsigned long));
  b = (unsigned long*) malloc (n * sizeof(unsigned long));
  c = (unsigned long*) malloc (2 * n * sizeof(unsigned long));
  stk = (unsigned long*) malloc (KarMulMem (n) * sizeof(unsigned long));

  for (i = 0; i < n; i++)
    {
      a[i] = (unsigned long) lrand48 ();
      b[i] = (unsigned long) lrand48 ();
    }

  if (!nokara)
    printf ("n=%u: ", n);

  st = cputime ();
  for (i = 0; i < N/n; i++)
      mul_basecase_n (c, a, b, n);
  printf ("mul took %dms", cputime () - st);
  
  if (nokara == 0)
    {
      st = cputime ();
      for (i = 0; i < N/n; i++)
	KarMul (c, a, b, n, stk);
      printf (", KarMul took %dms", cputime () - st);

      st = cputime ();
      for (i = 0; i < N/n; i++)
        ToomMul (c, a, b, n, stk);
      printf (", ToomMul took %dms", cputime () - st);
    }

  printf ("\n");
}
